#ifndef __CAFFEINE_H
#define __CAFFEINE_H

#include <stdio.h>

class Neon;
class Swirl;
class Growing;
class Mini;
class Circular;
class RotoPlasma;
class Invert;
class Cube;
class Glass;
class BumpMap;
class ShowText;

class ModFile;
class ModPlayer;

class Caffeine {
public:
	Caffeine(int width, int height, int bpp, int pitch);
	virtual ~Caffeine();

	void update(int curTime);
	void draw(unsigned char *buffer);

	void newPattern(int orderNr);

	virtual unsigned char *loadTitle(int *widthPtr, int *heightPtr, int *pitchPtr) = 0;
	virtual unsigned char *loadFont(int *widthPtr, int *heightPtr, int *pitchPtr) = 0;
	virtual unsigned char *loadBumpMap(int *widthPtr, int *heightPtr, int *pitchPtr, bool tileable) = 0;
	virtual unsigned char *loadText(int *widthPtr, int *heightPtr, int *pitchPtr) = 0;
	virtual void freeTitle() = 0;
	virtual void freeFont() = 0;
	virtual void freeBumpMap() = 0;
	virtual void freeText() = 0;
	virtual FILE *loadMod() = 0;

	virtual bool initAudio() = 0;
	virtual void startAudio(void (*callback)(int orderNr, void *data)) = 0;
	virtual void stopAudio() = 0;

	bool isFinished() { return finished; }

protected:
	int bpp;
private:
	int width, height, pitch;
	int part;
	int curTime, extTime, lastTime;
	int initStartTime;
	bool inited;
	int startTime;
	bool started;
	bool finished;

	bool modFade;
	int step, lastStep, lastStepTime;
	unsigned char *buffer2;
	int frames;

	bool bumpMapDone;

	unsigned char *title;
	unsigned char *font;
	unsigned char *bumpMap;

	Neon *neon;
	Swirl *swirl;
	Growing *growing;
	Mini *mini;
	Circular *circular;
	RotoPlasma *rotoplasma;
	Invert *invert;
	Cube *cube;
	Glass *glass;
	BumpMap *bumpmap;
	ShowText *showtext;

protected:
	ModFile *mod;
	ModPlayer *player;

private:
	void wipeInRight(unsigned char *buffer, unsigned char *title, int time, int duration);
	void wipeOutRight(unsigned char *buffer, unsigned char *title, int time, int duration);
	int wipeLength();
	int wipeFrameLength();
	int wipeSteps();
	void init();

	void copyLine32(unsigned int *out, unsigned int *in, int x);
	void copyLine16(unsigned short *out, unsigned short *in, int x);
	void drawLine32(unsigned int *out, int x);
	void drawLine16(unsigned short *out, int x);

	void drawDemo(unsigned char *buffer);
};


#endif
