#ifndef __WAVESTREAM_H__
#define __WAVESTREAM_H__

#include <e32base.h>
#include <mdaaudiooutputstream.h>
#include <mda/common/audio.h>

class MWaveGenerator {
public:
	virtual void RenderData(TAny* aUserData, TInt16* aPtr, TInt aLength) = 0;
	virtual void SetSampleRate(TInt aSampleRate) = 0;
};



enum TStreamCommand {
	EStreamStart, EStreamStop, EStreamSetVolume, EStreamKill
};

class CStreamShareData : public CBase {
public:
	CStreamShareData();

	RMutex iMutex;
	RMutex iAliveMutex;
	RSemaphore iCommandSemaphore;
	MWaveGenerator* iGen;
	TStreamCommand iCommand;
	TAny *iUserData;
	TRequestStatus* iStatusPtr;
	TBuf<100> iName;
	TBool iStopped;
	TInt iVolume;
};


class CWaveStream : public CBase {
public:
	static CWaveStream* NewL(const TDesC& aName, MWaveGenerator* aGen, TInt aVolume = 100);
	~CWaveStream();
	
	void Start();
	void Stop();
	void SetGenerator(MWaveGenerator* aGen);
	void SetUserData(TAny* aData);
	void LockShared();
	void UnlockShared();
	void SetVolume(TInt aVolume);
	
private:
	CWaveStream();
	void ConstructL(const TDesC& aName, MWaveGenerator* aGen, TInt aVolume);
	void SendCommand(TStreamCommand aCommand);

private:
	RThread iThread;
	CStreamShareData* iShare;

};


class CWaveThread : public CActive, public MMdaAudioOutputStreamCallback {
public:

	static TInt ThreadFunc(TAny* aPtr);
	static void ThreadFuncL(TAny* aPtr);
	static CWaveThread* NewL(CStreamShareData* aShare);

	void MaoscOpenComplete(TInt aError);
	void MaoscBufferCopied(TInt aError, const TDesC8& aBuffer);
	void MaoscPlayComplete(TInt aError);

	~CWaveThread();

	void DoCancel();
	void RunL();
	void StartRequest();

private:
	CWaveThread(CStreamShareData* aShare);
	void ConstructL();
	void RenderData(TPtr8 aPtr);
	void HandleCommand();

	void Start();
	void Stop();
	void UpdateVolume();

private:
	CStreamShareData* iShare;
	CMdaAudioOutputStream *iOutput;
	HBufC8* iBuf1;
	TMdaAudioDataSettings iSettings;
	TBool iStopping;
};


#endif
