#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bitmapfuncs.h"

void grayScale(unsigned int *ptr, int length) {
	unsigned int r,g,b;
	for (int i = 0; i < length; i++) {
		r = (*ptr&0x00FF0000)>>16;
		g = (*ptr&0x0000FF00)>>8;
		b = (*ptr&0x000000FF);
		*ptr++ = (r+g+b)/3;
	}
}

static void fadeFromBlack32(unsigned int *out, unsigned int *in, int width, int height, int pitch, int time, int duration) {
	unsigned int r,g,b;
	int fact = (1<<8)*time/duration;
	pitch /= 4;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			r = (((*in & 0x00FF0000) >> 16) * fact) >> 8;
			g = (((*in & 0x0000FF00) >> 8) * fact) >> 8;
			b = ((*in & 0x000000FF) * fact) >> 8;
			in++;
			*out++ = r<<16 | g<<8 | b;
		}
		in += pitch - width;
		out += pitch - width;
	}
}

static void fadeFromBlack16(unsigned short *out, unsigned short *in, int width, int height, int pitch, int time, int duration) {
	unsigned int r,g,b;
	int fact = (1<<8)*time/duration;
	pitch /= 2;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			r = (((*in & 0xF800) >> 11) * fact) >> 8;
			g = (((*in & 0x07E0) >> 5) * fact) >> 8;
			b = ((*in & 0x001F) * fact) >> 8;
			in++;
			*out++ = r<<11 | g<<5 | b;
		}
		in += pitch - width;
		out += pitch - width;
	}
}

static void fadeFromBlack12(unsigned short *out, unsigned short *in, int width, int height, int pitch, int time, int duration) {
	unsigned int r,g,b;
	int fact = (1<<8)*time/duration;
	pitch /= 2;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			r = (((*in & 0x0F00) >> 8) * fact) >> 8;
			g = (((*in & 0x00F0) >> 4) * fact) >> 8;
			b = ((*in & 0x000F) * fact) >> 8;
			in++;
			*out++ = r<<8 | g<<4 | b;
		}
		in += pitch - width;
		out += pitch - width;
	}
}

void fadeFromBlack(unsigned char *out, unsigned char *in, int width, int height, int pitch, int time, int duration, int bpp) {
	if (bpp == 12)
		fadeFromBlack12((unsigned short*) out, (unsigned short*) in, width, height, pitch, time, duration);
	else if (bpp == 16)
		fadeFromBlack16((unsigned short*) out, (unsigned short*) in, width, height, pitch, time, duration);
	else
		fadeFromBlack32((unsigned int*) out, (unsigned int*) in, width, height, pitch, time, duration);
}

static void fade32(unsigned int *out, unsigned int *in1, unsigned int *in2, int width, int height, int pitch, int time, int duration) {
	int r1,g1,b1,r2,g2,b2;
	unsigned int r,g,b;
	int fact = (1<<8)*time/duration;
	pitch /= 4;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			r1 = (*in1 & 0x00FF0000) >> 16;
			g1 = (*in1 & 0x0000FF00) >> 8;
			b1 = *in1 & 0x000000FF;
			r2 = (*in2 & 0x00FF0000) >> 16;
			g2 = (*in2 & 0x0000FF00) >> 8;
			b2 = *in2 & 0x000000FF;
			in1++;
			in2++;

			r = r1 + (((r2 - r1) * fact) >> 8);
			g = g1 + (((g2 - g1) * fact) >> 8);
			b = b1 + (((b2 - b1) * fact) >> 8);

			*out++ = r<<16 | g<<8 | b;
		}
		in1 += pitch - width;
		in2 += pitch - width;
		out += pitch - width;
	}
}

static void fade16(unsigned short *out, unsigned short *in1, unsigned short *in2, int width, int height, int pitch, int time, int duration) {
	int r1,g1,b1,r2,g2,b2;
	unsigned int r,g,b;
	int fact = (1<<8)*time/duration;
	pitch /= 2;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			r1 = (*in1 & 0xF800) >> 11;
			g1 = (*in1 & 0x07E0) >> 5;
			b1 = (*in1 & 0x001F);
			r2 = (*in2 & 0xF800) >> 11;
			g2 = (*in2 & 0x07E0) >> 5;
			b2 = (*in2 & 0x001F);
			in1++;
			in2++;

			r = r1 + (((r2 - r1) * fact) >> 8);
			g = g1 + (((g2 - g1) * fact) >> 8);
			b = b1 + (((b2 - b1) * fact) >> 8);

			*out++ = r<<11 | g<<5 | b;
		}
		in1 += pitch - width;
		in2 += pitch - width;
		out += pitch - width;
	}
}

static void fade12(unsigned short *out, unsigned short *in1, unsigned short *in2, int width, int height, int pitch, int time, int duration) {
	int r1,g1,b1,r2,g2,b2;
	unsigned char r,g,b;
	int fact = (1<<8)*time/duration;
	pitch /= 2;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			r1 = (*in1 & 0x0F00) >> 8;
			g1 = (*in1 & 0x00F0) >> 4;
			b1 = (*in1 & 0x000F);
			r2 = (*in2 & 0x0F00) >> 8;
			g2 = (*in2 & 0x00F0) >> 4;
			b2 = (*in2 & 0x000F);
			in1++;
			in2++;

			r = r1 + (((r2 - r1) * fact) >> 8);
			g = g1 + (((g2 - g1) * fact) >> 8);
			b = b1 + (((b2 - b1) * fact) >> 8);

			*out++ = r<<8 | g<<4 | b;
		}
		in1 += pitch - width;
		in2 += pitch - width;
		out += pitch - width;
	}
}

void fade(unsigned char *out, unsigned char *in1, unsigned char *in2, int width, int height, int pitch, int time, int duration, int bpp) {
	if (bpp == 12)
		fade12((unsigned short*) out, (unsigned short*) in1, (unsigned short*) in2, width, height, pitch, time, duration);
	else if (bpp == 16)
		fade16((unsigned short*) out, (unsigned short*) in1, (unsigned short*) in2, width, height, pitch, time, duration);
	else
		fade32((unsigned int*) out, (unsigned int*) in1, (unsigned int*) in2, width, height, pitch, time, duration);
}

void fadeToBlack(unsigned char *out, unsigned char *in, int width, int height, int pitch, int time, int duration, int bpp) {
	fadeFromBlack(out, in, width, height, pitch, duration - time, duration, bpp);
}

void drop(unsigned char *in1, unsigned char *in2, int width, int height, int pitch, int time, int duration, int bpp) {
	int y = height*time/duration;
	memcpy(in1, in2 + (height - y)*pitch, y*pitch);
}

void shrinkGray(unsigned char *out, int outwidth, int outheight, unsigned char *in, int inwidth, int inheight, int inpitch) {
	int inY = 0;
	int yInc = (1<<8)*inheight/outheight;
	int xInc = (1<<8)*inwidth/outwidth;
	unsigned int div = (1<<24)/(xInc*yInc);
	for (int y = 0; y < outheight; y++) {
		int inX = 0;
		int endY = inY + yInc;
		for (int x = 0; x < outwidth; x++) {
			int endX = inX + xInc;
			unsigned int sum = 0;
			unsigned char *outptr = out + outwidth*y + x;
			unsigned char *inptr;
			inptr = in + inpitch*(endY>>8) + (inX>>8);
			sum += *inptr++ * (0x100 - (inY & 0xFF)) * (0x100 - (inX & 0xFF));
			for (int ix = (inX>>8) + 1; ix < (endX>>8); ix++)
				sum += (*inptr++ << 8) * (0x100 - (inY & 0xFF));
			sum += *inptr * (0x100 - (inY & 0xFF)) * (endX & 0xFF);
			for (int iy = (inY>>8) + 1; iy < (endY>>8); iy++) {
				inptr = in + inpitch*iy + (inX>>8);
				sum += (*inptr++ << 8) * (0x100 - (inX & 0xFF));
				for (int ix = (inX>>8) + 1; ix < (endX>>8); ix++)
					sum += *inptr++ << 16;
				if (endX>>8 < inwidth)
					sum += (*inptr << 8) * (endX & 0xFF);
			}
			if (endY>>8 < inheight) {
				inptr = in + inpitch*(endY>>8) + (inX>>8);
				sum += *inptr++ * (endY & 0xFF) * (0x100 - (inX & 0xFF));
				for (int ix = (inX>>8) + 1; ix < (endX>>8); ix++)
					sum += (*inptr++ << 8) * (endY & 0xFF);
				if (endX>>8 < inwidth)
					sum += *inptr * (endY & 0xFF) * (endX & 0xFF);
			}
			sum = (sum * div) >> 24;
			*outptr = sum;
			inX += xInc;
		}
		inY += yInc;
	}
}

unsigned char* scaleTitle(unsigned char *titleScale, int titleWidth, int titleHeight, int titlePitch, int frameWidth, int frameHeight) {
	unsigned char *out = (unsigned char*) malloc(frameWidth*frameHeight);
	memset(out, 0, frameWidth*frameHeight);
	if (frameWidth < titleWidth) {
		int outframeHeight = titleHeight*frameWidth/titleWidth;
		int inframeHeight = titleHeight;
		if (outframeHeight > frameHeight) {
			outframeHeight = frameHeight;
			inframeHeight = frameHeight*titleWidth/frameWidth;
		}
		shrinkGray(out, frameWidth, outframeHeight, titleScale, titleWidth, inframeHeight, titlePitch);
	} else {
		int h = titleHeight;
		if (frameHeight < h)
			h = frameHeight;
		int xstart = (frameWidth - titleWidth)/2;
		for (int y = 0; y < h; y++)
			memcpy(out + y*frameWidth + xstart, titleScale + y*titlePitch, titleWidth);
	}
	return out;
}

unsigned char *tileCrop(unsigned char *in, int inwidth, int inheight, int inpitch, int frameWidth, int frameHeight) {
	int outw = frameWidth + 1;
	int outh = frameHeight + 1;
	unsigned char *out = (unsigned char*) malloc(outw*outh*sizeof(char));
	int w = inwidth;
	if (outw < w)
		w = outw;
	int h = inheight;
	if (outh < h)
		h = outh;
	for (int y = 0; y < h; y++) {
		memcpy(out + y*outw, in + y*inpitch, w*sizeof(char));
		for (int x = w; x < outw; x += inwidth) {
			int w2 = inwidth;
			if (outw - x < w2)
				w2 = outw - x;
			memcpy(out + y*outw + x, out + y*outw, w2*sizeof(char));
		}
	}
	for (int y = h; y < outh; y++)
		memcpy(out + y*outw, out + (y % inheight)*outw, outw*sizeof(char));
	return out;
}


void calcRotSize(float angle, int inwidth, int inheight, int *outwidth, int *outheight) {
	double cosa = cos(angle);
	double sina = sin(angle);
	int diag1W = abs((int)(inwidth*cosa - inheight*sina));
	int diag1H = abs((int)(inheight*cosa + inwidth*sina));
	int diag2W = abs((int)((-inwidth)*cosa - inheight*sina));
	int diag2H = abs((int)(inheight*cosa + (-inwidth)*sina));
	if (diag2W>diag1W) diag1W = diag2W;
	if (diag2H>diag1H) diag1H = diag2H;
	int outw = diag1W;
	int outh = diag1H;
	*outwidth = outw;
	*outheight = outh;
}

unsigned char *rotateGray(unsigned char *in, int inwidth, int inheight, int inpitch, float angle, int *outwidth, int *outheight) {
	double cosa = cos(angle);
	double sina = sin(angle);
	int outw, outh;
	calcRotSize(angle, inwidth, inheight, &outw, &outh);
	*outwidth = outw;
	*outheight = outh;

	unsigned char *out = (unsigned char*) malloc(outw*outh*sizeof(char));
	memset(out, 0xFF, outw*outh*sizeof(char));


	int corner1X = (int)((1<<16)*(inwidth  + (-outw)*cosa + (-outh)*sina)/2);
	int corner1Y = (int)((1<<16)*(inheight + (-outh)*cosa - (-outw)*sina)/2);
	int corner2X = (int)((1<<16)*(inwidth  +  (outw)*cosa + (-outh)*sina)/2);
	int corner2Y = (int)((1<<16)*(inheight + (-outh)*cosa -  (outw)*sina)/2);
	int corner3X = (int)((1<<16)*(inwidth  + (-outw)*cosa +  (outh)*sina)/2);
	int corner3Y = (int)((1<<16)*(inheight +  (outh)*cosa - (-outw)*sina)/2);

	int XdiffX = (corner2X - corner1X)/outw;
	int XdiffY = (corner2Y - corner1Y)/outw;
	int YdiffX = (corner3X - corner1X)/outh;
	int YdiffY = (corner3Y - corner1Y)/outh;

	unsigned char *outptr = out;

	for (int y = 0; y < outh; y++) {
		int posX = corner1X + y*YdiffX;
		int posY = corner1Y + y*YdiffY;
		for (int x = 0; x < outw; x++) {
			int ix = posX>>16;
			int iy = posY>>16;
			int xFrac = (posX>>8) & 0xFF;
			int yFrac = (posY>>8) & 0xFF;
			if (ix >= 0 && iy >= 0 && ix < inwidth-1 && iy < inheight-1) {
				int sum = 0;
				sum += in[(iy+0)*inpitch + ix+0] * (0x100 - yFrac) * (0x100 - xFrac);
				sum += in[(iy+0)*inpitch + ix+1] * (0x100 - yFrac) * (xFrac);
				sum += in[(iy+1)*inpitch + ix+0] * (yFrac) * (0x100 - xFrac);
				sum += in[(iy+1)*inpitch + ix+1] * (yFrac) * (xFrac);
				sum >>= 16;
				*outptr++ = sum;
			} else
				outptr++;

			posX += XdiffX;
			posY += XdiffY;
		}
	}

	return out;
}

void mergeImage(unsigned char *mask, int maskwidth, int maskheight, unsigned char *out, int outwidth, int outheight) {
	int offx = (outwidth - maskwidth)/2;
	int offy = (outheight - maskheight)/2;
	for (int y = 0; y < maskheight; y++) {
		unsigned char *outptr = out + (offy + y)*outwidth + offx;
		unsigned char *maskptr = mask + y*maskwidth;
		for (int x = 0; x < maskwidth; x++) {
			int val = *outptr;
			int mval = *maskptr++;
			val = (val*mval) >> 8;
			*outptr++ = val;
		}
	}
}

unsigned char *convertBitmap(unsigned char *in, int inbpp, int width, int height, int inpitch, int outbpp, bool release) {
	int bytes = (outbpp + 7)/8;
	if (inbpp == outbpp) {
		if (inpitch == bytes*width)
			return in;

		unsigned char *out = (unsigned char*) malloc(bytes*width*height);
		for (int y = 0; y < height; y++)
			memcpy(out + y*width*bytes, in + y*inpitch, bytes*width);
		if (release)
			free(in);
		return out;
	}

	unsigned char *out = (unsigned char*) malloc(bytes*width*height);
	if (inbpp == 8) {
		unsigned char *inptr = (unsigned char*) in;
		if (outbpp == 12) {
			unsigned short *outptr = (unsigned short*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int g = *inptr >> 4;
					inptr++;
					*outptr++ = (g << 8) | (g << 4) | (g);
				}
			}
		} else if (outbpp == 16) {
			unsigned short *outptr = (unsigned short*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = *inptr >> 3;
					int g = *inptr >> 2;
					inptr++;
					*outptr++ = (r << 11) | (g << 5) | (r);
				}
			}
		} else {
			unsigned int *outptr = (unsigned int*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int g = *inptr++;
					*outptr++ = (g << 16) | (g << 8) | (g);
				}
			}
		}
	} else if (inbpp == 12) {
		unsigned short *inptr = (unsigned short*) in;
		if (outbpp == 8) {
			unsigned char *outptr = (unsigned char*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int g = (*inptr & 0x00F0);
					g |= g >> 4;
					inptr++;
					*outptr++ = g;
				}
			}
		} else if (outbpp == 16) {
			unsigned short *outptr = (unsigned short*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = (*inptr & 0x0F00) >> 7;
					r |= r >> 4;
					int g = (*inptr & 0x00F0) >> 2;
					g |= g >> 4;
					int b = (*inptr & 0x000F) << 1;
					b |= b >> 4;
					inptr++;
					*outptr++ = (r << 11) | (g << 5) | (b);
				}
			}
		} else {
			unsigned int *outptr = (unsigned int*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = (*inptr & 0x0F00) >> 4;
					r |= r >> 4;
					int g = (*inptr & 0x00F0);
					g |= g >> 4;
					int b = (*inptr & 0x000F) << 4;
					b |= b >> 4;
					inptr++;
					*outptr++ = (r << 16) | (g << 8) | (b);
				}
			}
		}
	} else if (inbpp == 16) {
		unsigned short *inptr = (unsigned short*) in;
		if (outbpp == 8) {
			unsigned char *outptr = (unsigned char*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int g = (*inptr & 0x07E0) >> 3;
					g |= g >> 6;
					inptr++;
					*outptr++ = g;
				}
			}
		} else if (outbpp == 12) {
			unsigned short *outptr = (unsigned short*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = (*inptr & 0xF800) >> 12;
					int g = (*inptr & 0x07E0) >> 7;
					int b = (*inptr & 0x001F) >> 1;
					inptr++;
					*outptr++ = (r << 8) | (g << 4) | (b);
				}
			}
		} else {
			unsigned int *outptr = (unsigned int*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = (*inptr & 0xF800) >> 8;
					r |= r >> 5;
					int g = (*inptr & 0x07E0) >> 3;
					g |= g >> 6;
					int b = (*inptr & 0x001F) << 3;
					b |= b >> 5;
					inptr++;
					*outptr++ = (r << 16) | (g << 8) | (b);
				}
			}
		}
	} else {
		unsigned int *inptr = (unsigned int*) in;
		if (outbpp == 8) {
			unsigned char *outptr = (unsigned char*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int g = (*inptr & 0x00FF00) >> 8;
					inptr++;
					*outptr++ = g;
				}
			}
		} else if (outbpp == 12) {
			unsigned short *outptr = (unsigned short*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = (*inptr & 0xFF0000) >> 20;
					int g = (*inptr & 0x00FF00) >> 12;
					int b = (*inptr & 0x0000FF) >> 4;
					inptr++;
					*outptr++ = (r << 8) | (g << 4) | (b);
				}
			}
		} else {
			unsigned short *outptr = (unsigned short*) out;
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					int r = (*inptr & 0xFF0000) >> 19;
					int g = (*inptr & 0x00FF00) >> 10;
					int b = (*inptr & 0x0000FF) >> 3;
					inptr++;
					*outptr++ = (r << 11) | (g << 5) | (b);
				}
			}
		}
	}
	if (release)
		free(in);
	return out;
}


