#include <stdlib.h>
#include <math.h>
#include "bumpmap.h"

#define HYPOT(x,y) sqrt((x)*(x) + (y)*(y))

BumpMap::BumpMap(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
	if (w < h)
		zoomscale = 256*240/w;
	else
		zoomscale = 256*240/h;
	light = NULL;
}

BumpMap::~BumpMap(void) {
	free(light);
}

void BumpMap::init(void) {
	light = (unsigned char*) malloc(sizeof(char)*256*256);
	unsigned char *ptr = light;
	int dist;
	for (int y = 0; y < 256; y++) {
		for (int x = 0; x < 256; x++) {
			dist = (int) (2*HYPOT(x - 128, y - 128));
			if (dist>255)
				*ptr++ = 0;
			else
				*ptr++ = 255-dist;
		}
	}
}

static unsigned int shade12(unsigned int c, int intensity) {
	int r,g,b;
	r = (c & 0x0F00) >> 4;
	r |= r >> 4;
	g = c & 0x00F0;
	g |= g >> 4;
	b = (c & 0x000F) << 4;
	b |= b >> 4;
	r = (intensity*r)>>8;
	g = (intensity*g)>>8;
	b = (intensity*b)>>8;
	return ((r << 4) & 0x0F00) | (g & 0x00F0) | (b >> 4);
}

static unsigned int shade16(unsigned int c, int intensity) {
	int r,g,b;
	r = (c & 0xF800) >> 8;
	r |= r >> 5;
	g = (c & 0x07E0) >> 3;
	g |= g >> 6;
	b = (c & 0x001F) << 3;
	b |= b >> 5;
	r = (intensity*r)>>8;
	g = (intensity*g)>>8;
	b = (intensity*b)>>8;
	return ((r << 8) & 0xF800) | ((g << 3) & 0x07E0) | (b >> 3);
}

static unsigned int shade32(unsigned int c, int intensity) {
	int r,g,b;
	r = (c&0x00FF0000)>>16;
	g = (c&0x0000FF00)>>8;
	b = c&0x000000FF;
	r = (intensity*r)>>8;
	g = (intensity*g)>>8;
	b = (intensity*b)>>8;
	return r<<16 | g<<8 | b;
}

void BumpMap::draw(unsigned char *buffer, int time, unsigned char *image, unsigned char *map) {
	int mapWidth = width+1;

	int light1X = (int) (width/2+(width/2-15)*sin((float)(1300+time)/600));
	int light1Y = (int) (height/2+(height/2-15)*cos((float)(900+time)/450));
	int zoom1 = (int) (295+75*sin((float)time/800));
	int light2X = (int) (width/2+(width/2-25)*sin((float)(500+time)/750));
	int light2Y = (int) (height/2+(height/2-25)*sin((float)(1200+time)/600));
	int zoom2 = (int) (350+110*sin((float)time/2100));

	int zoom1inv = (1<<16) / zoom1;
	int zoom2inv = (1<<16) / zoom2;

	int intensity, posX, posY, diffX, diffY;
	if (bpp == 12) {
		unsigned short *imageptr = (unsigned short*) image;
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				intensity = 0;
				diffX = map[y*mapWidth+x] - map[y*mapWidth+x+1];
				diffY = map[y*mapWidth+x] - map[(y+1)*mapWidth+x];
				posX = (zoomscale*zoom1inv*(x-light1X+zoom1/2) >> 16) + diffX;
				posY = (zoomscale*zoom1inv*(y-light1Y+zoom1/2) >> 16) + diffY;
				if ((posX>=0) && (posX<256) && (posY>=0) && (posY<256))
					intensity += light[(posY<<8)+posX];
				posX = (zoomscale*zoom2inv*(x-light2X+zoom2/2) >> 16) + diffX;
				posY = (zoomscale*zoom2inv*(y-light2Y+zoom2/2) >> 16) + diffY;
				if ((posX>=0) && (posX<256) && (posY>=0) && (posY<256))
					intensity += light[(posY<<8)+posX];
				if (intensity>256) intensity = 256;
				*outptr++ = shade12(*imageptr++,intensity);
			}
			imageptr += spitch - width;
			outptr += spitch - width;
		}
	} else if (bpp == 16) {
		unsigned short *imageptr = (unsigned short*) image;
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				intensity = 0;
				diffX = map[y*mapWidth+x] - map[y*mapWidth+x+1];
				diffY = map[y*mapWidth+x] - map[(y+1)*mapWidth+x];
				posX = (zoomscale*zoom1inv*(x-light1X+zoom1/2) >> 16) + diffX;
				posY = (zoomscale*zoom1inv*(y-light1Y+zoom1/2) >> 16) + diffY;
				if ((posX>=0) && (posX<256) && (posY>=0) && (posY<256))
					intensity += light[(posY<<8)+posX];
				posX = (zoomscale*zoom2inv*(x-light2X+zoom2/2) >> 16) + diffX;
				posY = (zoomscale*zoom2inv*(y-light2Y+zoom2/2) >> 16) + diffY;
				if ((posX>=0) && (posX<256) && (posY>=0) && (posY<256))
					intensity += light[(posY<<8)+posX];
				if (intensity>256) intensity = 256;
				*outptr++ = shade16(*imageptr++,intensity);
			}
			imageptr += spitch - width;
			outptr += spitch - width;
		}
	} else {
		unsigned int *imageptr = (unsigned int*) image;
		unsigned int *outptr = (unsigned int*) buffer;
		int ipitch = pitch/4;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				intensity = 0;
				diffX = map[y*mapWidth+x] - map[y*mapWidth+x+1];
				diffY = map[y*mapWidth+x] - map[(y+1)*mapWidth+x];
				posX = (zoomscale*zoom1inv*(x-light1X+zoom1/2) >> 16) + diffX;
				posY = (zoomscale*zoom1inv*(y-light1Y+zoom1/2) >> 16) + diffY;
				if ((posX>=0) && (posX<256) && (posY>=0) && (posY<256))
					intensity += light[(posY<<8)+posX];
				posX = (zoomscale*zoom2inv*(x-light2X+zoom2/2) >> 16) + diffX;
				posY = (zoomscale*zoom2inv*(y-light2Y+zoom2/2) >> 16) + diffY;
				if ((posX>=0) && (posX<256) && (posY>=0) && (posY<256))
					intensity += light[(posY<<8)+posX];
				if (intensity>256) intensity = 256;
				*outptr++ = shade32(*imageptr++,intensity);
			}
			imageptr += ipitch - width;
			outptr += ipitch - width;
		}
	}

}

