#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "caffeine.h"
#include "modfile.h"
#include "modplayer.h"

#include "neon.h"
#include "swirl.h"
#include "growing.h"
#include "mini.h"
#include "circular.h"
#include "rotoplasma.h"
#include "invert.h"
#include "cube.h"
#include "glass.h"
#include "bumpmap.h"
#include "showtext.h"

#include "bitmapfuncs.h"

static void staticCallback(int orderNr, void *data) {
	Caffeine *c = (Caffeine*) data;
	c->newPattern(orderNr);
}

void Caffeine::newPattern(int orderNr) {
	part++;
	lastTime = curTime;
}

void Caffeine::copyLine32(unsigned int *out, unsigned int *in, int x) {
	out += x;
	in += x;
	int ipitch = pitch/4;
	for (int i = 0; i < height; i++) {
		*out = *in;
		out += ipitch;
		in += width;
	}
}

void Caffeine::copyLine16(unsigned short *out, unsigned short *in, int x) {
	out += x;
	in += x;
	int spitch = pitch/2;
	for (int i = 0; i < height; i++) {
		*out = *in;
		out += spitch;
		in += width;
	}
}

int Caffeine::wipeLength() {
	int stepWidth = (int) ceil((float)width/wipeSteps());
	int numFrames = stepWidth + wipeSteps() - 1;
	return numFrames;
}

int Caffeine::wipeFrameLength() {
	return 45;
}

int Caffeine::wipeSteps() {
	return 8;
}

void Caffeine::wipeInRight(unsigned char *buffer, unsigned char *title, int time, int duration) {
	int step = wipeSteps();

	int i = time*wipeLength()/duration;
	int bytes = (bpp + 7)/8;
	int lastx = (i + 1)*step - 1 - (step + 1)*(step - 1);

	if (lastx <= 0) {
		memset(buffer, 0, pitch*height);
	} else {
		for (int y = 0; y < height; y++) {
			memcpy(buffer + y*pitch, title + y*bytes*width, (lastx-1)*bytes);
			if (lastx + 1 < width)
				memset(buffer + y*pitch + (lastx+1)*bytes, 0, bytes*(width - (lastx + 1)));
		}
	}

	if (bpp <= 16) {
		for (int j = 0; j < step; j++) {
			for (int k = 0; k <= j; k++) {
				int x = (i + 1)*step - 1 - step*j - k;
				if (x >= 0 && x < width)
					copyLine16((unsigned short*) buffer, (unsigned short*) title, x);
			}
		}
	} else {
		for (int j = 0; j < step; j++) {
			for (int k = 0; k <= j; k++) {
				int x = (i + 1)*step - 1 - step*j - k;
				if (x >= 0 && x < width)
					copyLine32((unsigned int*) buffer, (unsigned int*) title, x);
			}
		}
	}
}

void Caffeine::drawLine32(unsigned int *out, int x) {
	out += x;
	int ipitch = pitch/4;
	for (int i = 0; i < height; i++) {
		*out = 0;
		out += ipitch;
	}
}

void Caffeine::drawLine16(unsigned short *out, int x) {
	out += x;
	int spitch = pitch/2;
	for (int i = 0; i < height; i++) {
		*out = 0;
		out += spitch;
	}
}

static void copyBitmap(unsigned char *out, int width, int height, int bpp, int outpitch, unsigned char *in, int inpitch) {
	int bytes = (bpp + 7)/8;
	if (outpitch == inpitch) {
		memcpy(out, in, inpitch*height);
	} else {
		for (int y = 0; y < height; y++)	
			memcpy(out + outpitch*y, in + inpitch*y, bytes*width);
	}
}

void Caffeine::wipeOutRight(unsigned char *buffer, unsigned char *title, int time, int duration) {
	int step = wipeSteps();

	int i = time*wipeLength()/duration;
	int bytes = (bpp + 7)/8;
	int lastx = (i + 1)*step - 1 - (step + 1)*(step - 1);

	if (lastx <= 0) {
		copyBitmap(buffer, width, height, bpp, pitch, title, bytes*width);
	} else {
		for (int y = 0; y < height; y++) {
			memset(buffer + y*pitch, 0, bytes*(lastx - 1));
			if (lastx + 1 < width)
				memcpy(buffer + y*pitch + (lastx + 1)*bytes, title + y*bytes*width + (lastx + 1)*bytes, bytes*(width - (lastx + 1)));
		}
	}

	if (bpp <= 16) {
		for (int j = 0; j < step; j++) {
			for (int k = 0; k <= j; k++) {
				int x = (i + 1)*step - 1 - step*j - k;
				if (x >= 0 && x < width)
					drawLine16((unsigned short*) buffer, x);
			}
		}
	} else {
		for (int j = 0; j < step; j++) {
			for (int k = 0; k <= j; k++) {
				int x = (i + 1)*step - 1 - step*j - k;
				if (x >= 0 && x < width)
					drawLine32((unsigned int*) buffer, x);
			}
		}
	}
}

Caffeine::Caffeine(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;

	inited = false;
	started = false;
	finished = false;
	startTime = -1;
	initStartTime = -1;

	neon = NULL;
	swirl = NULL;
	growing = NULL;
	mini = NULL;
	circular = NULL;
	rotoplasma = NULL;
	invert = NULL;
	cube = NULL;
	glass = NULL;
	bumpmap = NULL;
	showtext = NULL;

	font = NULL;
	title = NULL;
	bumpMap = NULL;
	mod = NULL;
	player = NULL;
	buffer2 = NULL;
}

Caffeine::~Caffeine() {
	delete neon;
	delete swirl;
	delete growing;
	delete mini;
	delete circular;
	delete rotoplasma;
	delete invert;
	delete cube;
	delete glass;
	delete bumpmap;
	delete showtext;
	free(font);
	free(title);
	free(bumpMap);
	delete mod;
	delete player;
	free(buffer2);
}

void Caffeine::update(int extTime) {
	if (startTime == -1)
		startTime = extTime;

	this->extTime = extTime;
	curTime = extTime - startTime;

	int wipeTime = wipeLength()*wipeFrameLength();
	if (!inited) {
		if (title == NULL) {
			int titleWidth = 0, titleHeight = 0, titlePitch = 0;
			unsigned char *titleScale = loadTitle(&titleWidth, &titleHeight, &titlePitch);
			title = scaleTitle(titleScale, titleWidth, titleHeight, titlePitch, width, height);
			freeTitle();
			title = convertBitmap(title, 8, width, height, width, bpp);
		}

		if (curTime >= wipeTime) {
			init();
		}
	} else if (!started) {
		if (curTime >= wipeTime) {
			started = true;
			startTime = extTime;

			part = -1;
			modFade = false;
			step = -1;
			lastStep = 0;
			frames = 0;
			curTime = 0;

			startAudio(staticCallback);
		}
	}
}

void Caffeine::draw(unsigned char *buffer) {
	int bytes = (bpp + 7)/8;

	int wipeTime = wipeLength()*wipeFrameLength();
	if (!inited) {
		if (curTime < wipeTime) {
			wipeInRight(buffer, title, curTime, wipeTime);
		} else {
			copyBitmap(buffer, width, height, bpp, pitch, title, bytes*width);
		}
	} else if (!started) {
		if (curTime < 0) {
			copyBitmap(buffer, width, height, bpp, pitch, title, bytes*width);
		} else if (curTime < wipeTime) {
			wipeOutRight(buffer, title, curTime, wipeTime);
		} else {
			memset(buffer, 0, pitch*height);
		}
	} else {
		drawDemo(buffer);
	}
}

void Caffeine::init() {
	if (initStartTime == -1)
		initStartTime = curTime;

	if (font == NULL) {
		int fontWidth = 0, fontHeight = 0, fontPitch = 0;
		font = loadFont(&fontWidth, &fontHeight, &fontPitch);

		if (fontWidth != 160 || fontHeight != 112) {
			fprintf(stderr, "Bad font file\n");
			finished = true;
			return;
		}
	}

	if (bumpMap == NULL) {
		int bumpMapWidth = 0, bumpMapHeight = 0, bumpMapPitch = 0;
		if (width > 320 || height > 240)
			bumpMap = loadBumpMap(&bumpMapWidth, &bumpMapHeight, &bumpMapPitch, true);
		else
			bumpMap = loadBumpMap(&bumpMapWidth, &bumpMapHeight, &bumpMapPitch, false);

		if (bumpMap == NULL) {
			fprintf(stderr, "Bad bumpmap file\n");
			finished = true;
			return;
		}

		if (width != bumpMapWidth - 1 || height != bumpMapHeight - 1) {
			unsigned char *tmp = tileCrop(bumpMap, bumpMapWidth, bumpMapHeight, bumpMapPitch, width, height);
			freeBumpMap();
			bumpMap = tmp;
		} else {
			unsigned char *tmp = (unsigned char*) malloc(sizeof(char)*bumpMapWidth*bumpMapHeight);
			copyBitmap(tmp, bumpMapWidth, bumpMapHeight, 8, bumpMapWidth, bumpMap, bumpMapPitch);
			freeBumpMap();
			bumpMap = tmp;
		}

		bumpMapDone = false;
		return;
	}

	if (!bumpMapDone) {
		int textWidth = 0, textHeight = 0, textPitch = 0;
		unsigned char *textImage = loadText(&textWidth, &textHeight, &textPitch);

		int rotTextWidth = 0, rotTextHeight = 0;
		float angle = -atan2(height, width);
		calcRotSize(angle, textWidth, textHeight, &rotTextWidth, &rotTextHeight);

		bool origFreed = false;

		int shrinkRotTextWidth = width*9/10;
		if (shrinkRotTextWidth < rotTextWidth) {
			int shrinkTextWidth = textWidth*shrinkRotTextWidth/rotTextWidth;
			int shrinkTextHeight = textHeight*shrinkTextWidth/textWidth;
			unsigned char *tmp = (unsigned char*) malloc(shrinkTextWidth*shrinkTextHeight*sizeof(char));
			shrinkGray(tmp, shrinkTextWidth, shrinkTextHeight, textImage, textWidth, textHeight, textPitch);
			freeText();
			origFreed = true;
			textImage = tmp;
			textWidth = shrinkTextWidth;
			textHeight = shrinkTextHeight;
			textPitch = textWidth;
		}

		unsigned char *rotTextImage = rotateGray(textImage, textWidth, textHeight, textPitch, -atan2(height, width), &rotTextWidth, &rotTextHeight);

		if (origFreed)
			free(textImage);
		else
			freeText();

		mergeImage(rotTextImage, rotTextWidth, rotTextHeight, bumpMap, width + 1, height + 1);
		free(rotTextImage);

		bumpMapDone = true;
		return;
	}

	if (mod == NULL) {
		FILE *in = loadMod();
		if (in == NULL) {
			finished = true;
			return;
		}

		mod = new ModFile(in);
		fclose(in);

		if (!initAudio()) {
			finished = true;
			return;
		}

		return;
	}

	if (neon == NULL) {
		neon = new Neon(width, height, bpp, pitch);
		neon->init();
		return;
	}
	if (swirl == NULL) {
		swirl = new Swirl(width, height, bpp, pitch, 1250);
		swirl->init();
		return;
	}
	if (growing == NULL) {
		growing = new Growing(width, height, bpp, pitch);
		growing->init();
		return;
	}
	if (mini == NULL) {
		mini = new Mini(width, height, bpp, pitch);
		mini->init();
		return;
	}
	if (circular == NULL) {
		circular = new Circular(width, height, bpp, pitch);
		circular->init(neon->getDisttable());
		return;
	}
	if (rotoplasma == NULL) {
		rotoplasma = new RotoPlasma(width, height, bpp, pitch);
	}
	if (!rotoplasma->isInited()) {
		rotoplasma->init();
		return;
	}
	if (invert == NULL) {
		invert = new Invert(width, height, bpp, pitch);
		invert->init();
		return;
	}
	if (cube == NULL) {
		cube = new Cube(width, height, bpp, pitch, 0x00FF0000);
		cube->init();
		return;
	}
	if (glass == NULL) {
		glass = new Glass(width, height, bpp, pitch);
		glass->init();
		return;
	}
	if (bumpmap == NULL) {
		bumpmap = new BumpMap(width, height, bpp, pitch);
		bumpmap->init();
		return;
	}
	if (showtext == NULL) {
		showtext = new ShowText(width, height, bpp, pitch);
		showtext->init();
		showtext->setFont(font, " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]", false, 16, 16, 10, true, 0x00FF0000);
		char *text1[] = {
			"Caffeine",
			"Code & Graphics\n\nwbs",
			"Music\n\nThe P",
			"Thanks to all\nmy friends.\n\nYou know who\nyou are.",
			"Greets to\neverybody at\nAssembly 06!",
			NULL
		};
		char *text2[] = {
			"Caffeine",
			"Code\n&\nGraphics\n\nwbs",
			"Music\n\nThe P",
			"Thanks to\nall my\nfriends.\n\nYou know\nwho\nyou are.",
			"Greets to\neverybody\nat\nAssembly\n06!",
			NULL
		};
		char **text = text1;
		if (width < 270)
			text = text2;
		showtext->setText(text, 17500);
	}
	if (buffer2 == NULL) {
		buffer2 = (unsigned char*) malloc(pitch*height);
		return;
	}

	if (curTime > initStartTime + 2500) {
		inited = true;
		startTime = extTime;
	}

}




void Caffeine::drawDemo(unsigned char *buffer) {
	if (part < 0) {
		memset(buffer, 0, pitch*height);
		return;
	}

	frames++;

	if (part < 2) {
		if (step <= 0 && curTime - lastTime < 500) {
			step = 0;
			neon->draw(buffer, curTime);
			fadeFromBlack(buffer, buffer, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 1;
			neon->draw(buffer, curTime);
		}
	} else if (part < 4) {
		if (step <= 2 && curTime - lastTime < 750) {
			step = 2;
			neon->draw(buffer, curTime);
			swirl->draw(buffer2, curTime);
			growing->draw(buffer, buffer, buffer2, curTime - lastTime, 750);
		} else {
			step = 3;
			swirl->draw(buffer, curTime);
		}
	} else if (part < 5) {
		if (step <= 4 && curTime - lastTime < 500) {
			step = 4;
			swirl->draw(buffer, curTime);
			memcpy(buffer2, buffer, pitch*height);
			mini->draw(buffer2);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 5;
			swirl->draw(buffer, curTime);
			mini->draw(buffer);
		}
	} else if (part < 7) {
		if (step <= 6 && curTime - lastTime < 500) {
			step = 6;
			swirl->draw(buffer, curTime);
			mini->draw(buffer);
			circular->draw(buffer2, curTime);
			drop(buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 7;
			circular->draw(buffer, curTime);
		}
	} else if (part < 10) {
		if (step <= 8 && curTime - lastTime < 500) {
			step = 8;
			circular->draw(buffer, curTime);
			rotoplasma->draw(buffer2, curTime);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 9;
			rotoplasma->draw(buffer, curTime);
		}
	} else if (part < 12) {
		if (step <= 10 && curTime - lastTime < 500) {
			step = 10;
			rotoplasma->draw(buffer, curTime);
			neon->draw(buffer2, curTime);
			invert->draw(buffer2, curTime);
			mini->draw(buffer2);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			neon->draw(buffer, curTime);
			invert->draw(buffer, curTime);
			mini->draw(buffer);
			step = 11;
		}
	} else if (part < 14) {
		if (step <= 12 && curTime - lastTime < 500) {
			step = 12;
			neon->draw(buffer, curTime);
			circular->draw(buffer2, curTime);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
			invert->draw(buffer, curTime);
			mini->draw(buffer);
		} else {
			step = 13;
			circular->draw(buffer, curTime);
			invert->draw(buffer, curTime);
			mini->draw(buffer);
		}
	} else if (part < 15) {
		if (step <= 14 && curTime - lastTime < 500) {
			step = 14;
			circular->draw(buffer, curTime);
			invert->draw(buffer, curTime);
			mini->draw(buffer);
			neon->draw(buffer2, curTime);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 15;
			neon->draw(buffer, curTime);
		}
	} else if (part < 16) {
		if (step <= 16 && curTime - lastTime < 500) {
			step = 16;
			neon->draw(buffer, curTime);
			memcpy(buffer2, buffer, pitch*height);
			cube->draw(buffer2, curTime, false);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 17;
			neon->draw(buffer, curTime);
			cube->draw(buffer, curTime, false);
		}
	} else if (part < 18) {
		if (step <= 18 && curTime - lastTime < 500) {
			step = 18;
			neon->draw(buffer, curTime);
			glass->draw(buffer2, curTime, buffer, bumpMap);
			cube->draw(buffer, curTime, false);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 19;
			neon->draw(buffer, curTime);
			glass->draw(buffer, curTime, buffer, bumpMap);
		}
	} else if (part < 20) {
		if (step <= 20 && curTime - lastTime < 500) {
			step = 20;
			neon->draw(buffer, curTime);
			bumpmap->draw(buffer2, curTime, buffer, bumpMap);
			glass->draw(buffer, curTime, buffer, bumpMap);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 21;
			neon->draw(buffer, curTime);
			bumpmap->draw(buffer, curTime, buffer, bumpMap);
		}
	} else if (part < 25) {
		if (step <= 22 && curTime - lastTime < 500) {
			step = 22;
			neon->draw(buffer2, curTime);
			bumpmap->draw(buffer, curTime, buffer2, bumpMap);
			fade(buffer, buffer, buffer2, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			step = 23;
			neon->draw(buffer, curTime);
			showtext->draw(buffer, curTime);
		}
	} else if (part < 26) {
		if (step <= 24 && curTime - lastTime < 500) {
			step = 24;
			if (!modFade) {
				player->fadeOut(500);
				modFade = true;
			}
			neon->draw(buffer, curTime);
			fadeToBlack(buffer, buffer, width, height, pitch, curTime - lastTime, 500, bpp);
		} else {
			memset(buffer, 0, pitch*height);
			finished = true;
		}
	} else {
		finished = true;
	}

	if (step != lastStep) {
//		if (frames>1)
//			printf("%d: %d fps\n",lastStep,1000*(frames-1)/(curTime-lastStepTime));
		frames = 1;
		lastStep = step;
//		lastStepTime = curTime;
	}

}

