#include <stdlib.h>
#include <math.h>
#include "circular.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define HYPOT(x,y) sqrt((x)*(x) + (y)*(y))

Circular::Circular(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
	sintable = NULL;
	disttable = NULL;
}

Circular::~Circular(void) {
	free(sintable);
	if (owntable)
		free(disttable);
}

void Circular::init(unsigned int *table) {
	int i;
//	xvar = (512-width)/2;
//	yvar = (512-height)/2;
	xvar = (512 - 320)/2;
	yvar = (512 - 240)/2;

	halfsizeX = 256*width/320;
	halfsizeY = 256*height/240;
	distsizeX = 2*halfsizeX;
	distsizeY = 2*halfsizeY;

	float distscaling = 5;
	if (width > height)
		distscaling = 5.*320/width;
	else
		distscaling = 5.*240/height;

	sintable = (unsigned int*) malloc(sizeof(int)*512);
	for (i = 0; i < 512; i++)
		sintable[i] = (int) (128+127*sin(2*M_PI*i/512));

	if (!table) {
		owntable = true;
		disttable = (unsigned int*) malloc(sizeof(int)*distsizeX*distsizeY);
		unsigned int *ptr = disttable;
		for (int y = 0; y < distsizeY; y++)
			for (int x = 0; x < distsizeX; x++)
				*ptr++ = (int) ((1<<8)*distscaling*HYPOT(x - halfsizeX, y - halfsizeY));
	} else {
		disttable = table;
		owntable = false;
	}


	unsigned int *palette = createPalette();
	for (i = 0; i < 512; i++)
		sintable[i] = palette[sintable[i]];
	free(palette);
}

unsigned int *Circular::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	for (int i = 0; i < 256; i++)
		*ptr++ = i<<16 | i<<8;

	if (bpp == 12) {
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>12) & 0x0F00) | ((pal[i]>>8) & 0x00F0) | ((pal[i]>>4) & 0x000F);
	} else if (bpp == 16) {
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>8) & 0xF800) | ((pal[i]>>5) & 0x07E0) | ((pal[i]>>3) & 0x001F);
	}

	return pal;
}

void Circular::draw(unsigned char *buffer, int time) {
	int offsetX1 = (int) (xvar+xvar*sin((double)(1950+time)/500));
	int offsetY1 = (int) (yvar+yvar*cos((double)(4650+time)/900));
	int offsetX2 = (int) (xvar+xvar*sin((double)(6200+time)/650));
	int offsetY2 = (int) (yvar+yvar*cos((double)(1250+time)/550));

	int noffsetX1 = halfsizeX - (256 - offsetX1)*width/320;
	int noffsetX2 = halfsizeX - (256 - offsetX2)*width/320;
	int noffsetY1 = halfsizeY - (256 - offsetY1)*height/240;
	int noffsetY2 = halfsizeY - (256 - offsetY2)*height/240;

	int size = (int) ((1<<8)*(15+5*sin((double)(1200+time)/700)));
	size /= 5;

	int dist1, dist2, c1, c2;
	unsigned int *distptr1, *distptr2;
	if (bpp <= 16) {
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			distptr1 = disttable + ((y + noffsetY1)*distsizeX) + noffsetX1;
			distptr2 = disttable + ((y + noffsetY2)*distsizeX) + noffsetX2;
			for (int x = 0; x < width; x++) {
				dist1 = *distptr1++;
				dist2 = *distptr2++;
				c1 = sintable[((dist1*size)>>16)&0x01FF];
				c2 = sintable[((dist2*size)>>16)&0x01FF];
				if (dist1>dist2)
					*outptr++ = (c1>c2)?c1:c2;
				else
					*outptr++ = (c1<c2)?c1:c2;
			}
			outptr += spitch - width;
		}
	} else {
		unsigned int *outptr = (unsigned int*) buffer;
		int ipitch = pitch/4;
		for (int y = 0; y < height; y++) {
			distptr1 = disttable + ((y + noffsetY1)*distsizeX) + noffsetX1;
			distptr2 = disttable + ((y + noffsetY2)*distsizeX) + noffsetX2;
			for (int x = 0; x < width; x++) {
				dist1 = *distptr1++;
				dist2 = *distptr2++;
				c1 = sintable[((dist1*size)>>16)&0x01FF];
				c2 = sintable[((dist2*size)>>16)&0x01FF];
				if (dist1>dist2)
					*outptr++ = (c1>c2)?c1:c2;
				else
					*outptr++ = (c1<c2)?c1:c2;
			}
			outptr += ipitch - width;
		}
	}
}


