#include <math.h>
#include "glass.h"

Glass::Glass(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
}

Glass::~Glass(void) {
}

void Glass::init(void) {
}

void Glass::draw(unsigned char *buffer, int time, unsigned char *image, unsigned char *map) {
	int r,g,b;
	int pixel, diff;
	int mapWidth = width+1;
	if (bpp == 12) {
		unsigned short *imageptr = (unsigned short*) image;
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				pixel = *imageptr++;
				diff = map[y*mapWidth+x]-map[y*mapWidth+x+1]+map[y*mapWidth+x]-map[(y+1)*mapWidth+x];
				r = (pixel & 0x0F00) >> 4;
				r |= r >> 4;
				g = pixel & 0x00F0;
				g |= g >> 4;
				b = (pixel & 0x000F) << 4;
				b |= b >> 4;
				r += diff;
				g += diff;
				b += diff;
				if (r<0) r = 0;
				else if (r>255) r = 255;
				if (g<0) g = 0;
				else if (g>255) g = 255;
				if (b<0) b = 0;
				else if (b>255) b = 255;
				*outptr++ = ((r << 4) & 0x0F00) | (g & 0x00F0) | (b >> 4);
			}
			imageptr += spitch - width;
			outptr += spitch - width;
		}
	} else if (bpp == 16) {
		unsigned short *imageptr = (unsigned short*) image;
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				pixel = *imageptr++;
				diff = map[y*mapWidth+x]-map[y*mapWidth+x+1]+map[y*mapWidth+x]-map[(y+1)*mapWidth+x];
				r = (pixel & 0xF800) >> 8;
				r |= r >> 5;
				g = (pixel & 0x07E0) >> 3;
				g |= g >> 6;
				b = (pixel & 0x001F) << 3;
				b |= b >> 5;
				r += diff;
				g += diff;
				b += diff;
				if (r<0) r = 0;
				else if (r>255) r = 255;
				if (g<0) g = 0;
				else if (g>255) g = 255;
				if (b<0) b = 0;
				else if (b>255) b = 255;
				*outptr++ = ((r << 8) & 0xF800) | ((g << 3) & 0x07E0) | (b >> 3);
			}
			imageptr += spitch - width;
			outptr += spitch - width;
		}
	} else {
		unsigned int *imageptr = (unsigned int*) image;
		unsigned int *outptr = (unsigned int*) buffer;
		int ipitch = pitch/4;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				pixel = *imageptr++;
				diff = map[y*mapWidth+x]-map[y*mapWidth+x+1]+map[y*mapWidth+x]-map[(y+1)*mapWidth+x];
				r = (pixel&0x00FF0000)>>16;
				g = (pixel&0x0000FF00)>>8;
				b = pixel&0x000000FF;
				r += diff;
				g += diff;
				b += diff;
				if (r<0) r = 0;
				else if (r>255) r = 255;
				if (g<0) g = 0;
				else if (g>255) g = 255;
				if (b<0) b = 0;
				else if (b>255) b = 255;
				*outptr++ = r<<16 | g<<8 | b;
			}
			imageptr += ipitch - width;
			outptr += ipitch - width;
		}
	}
}

