#include <stdlib.h>
#include <math.h>
#include "growing.h"

#define HYPOT(x,y) sqrt((x)*(x) + (y)*(y))

Growing::Growing(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
	disttable = NULL;
}

Growing::~Growing(void) {
	free(disttable);
}

void Growing::init(void) {
	disttable = (int*) malloc(sizeof(int)*width*height);
	float fpointx[] = { 0.42, 0.78, 0.71 };
	float fpointy[] = { 0.63, 0.21, 0.83 };
	int pointx[3];
	int pointy[3];
	int extra[3] = {0,20,55};
	if (width > height) {
		for (int i = 0; i < 3; i++) {
			pointx[i] = (int)(fpointx[i]*width);
			pointy[i] = (int)(fpointy[i]*height);
			extra[i] = extra[i]*height/240;
		}
	} else {
		for (int i = 0; i < 3; i++) {
			pointy[i] = (int)(fpointx[i]*height);
			pointx[i] = (int)((1 - fpointy[i])*width);
			extra[i] = extra[i]*width/240;
		}
	}

	int *ptr = disttable;
	maxDist = 0;
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			int min = (int)((1<<8)*(HYPOT(pointx[0] - x, pointy[0] - y) + extra[0]));
			for (int i = 1; i < 3; i++) {
				int d = (int)((1<<8)*(HYPOT(pointx[i] - x, pointy[i] - y) + extra[i]));
				if (d < min) min = d;
			}
			if (min > maxDist) maxDist = min;
			*ptr++ = min;
		}
	}
}

void Growing::draw(unsigned char *outbuf, unsigned char *in1buf, unsigned char *in2buf, int time, int duration) {
	int pos = maxDist*time/duration;
	int *distptr = disttable;
	if (bpp <= 16) {
		unsigned short *out = (unsigned short*) outbuf;
		unsigned short *in1 = (unsigned short*) in1buf;
		unsigned short *in2 = (unsigned short*) in2buf;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				if ((*distptr>>9)==(pos>>9)) *out = 0;
				else if (*distptr<pos) *out = *in2;
				else *out = *in1;
				distptr++;
				out++;
				in1++;
				in2++;
			}
			out += spitch - width;
			in1 += spitch - width;
			in2 += spitch - width;
		}
	} else {
		unsigned int *out = (unsigned int*) outbuf;
		unsigned int *in1 = (unsigned int*) in1buf;
		unsigned int *in2 = (unsigned int*) in2buf;
		int ipitch = pitch/4;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				if ((*distptr>>9)==(pos>>9)) *out = 0;
				else if (*distptr<pos) *out = *in2;
				else *out = *in1;
				distptr++;
				out++;
				in1++;
				in2++;
			}
			out += ipitch - width;
			in1 += ipitch - width;
			in2 += ipitch - width;
		}
	}
}


