#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <math.h>
#include <string.h>
#include "loadpng.h"

#include "caffeine.h"
#include "bitmapfuncs.h"

#include "audio.h"
#include "thread.h"

#ifdef __APPLE__
#include <Carbon/Carbon.h>
#include <unistd.h>
#endif


//#define WIDTH 176
//#define HEIGHT 208
//#define WIDTH 320
//#define HEIGHT 240
#define WIDTH 356
#define HEIGHT 416
//#define WIDTH 800
//#define HEIGHT 600
#define BPP 16


SDL_Surface *screen;
SDL_Surface *sdl_buffer;

unsigned char *lock() {
	if (SDL_MUSTLOCK(sdl_buffer)) SDL_LockSurface(sdl_buffer);
	return (unsigned char*) sdl_buffer->pixels;
}

void unlockFlip() {
	if (SDL_MUSTLOCK(sdl_buffer)) SDL_UnlockSurface(sdl_buffer);
	SDL_BlitSurface(sdl_buffer,NULL,screen,NULL);
	SDL_Flip(screen);
}

unsigned char *loadPng(const char *name, int *widthPtr, int *heightPtr, int *pitchPtr) {
	FILE *in = fopen(name, "rb");
	if (in == NULL) {
		perror(name);
		return NULL;
	}
	unsigned char *bitmap = (unsigned char*) loadPng(in, widthPtr, heightPtr, SDL_BYTEORDER == SDL_LIL_ENDIAN);
	*pitchPtr = *widthPtr*4;
	fclose(in);
	return bitmap;
}

unsigned char *loadPng8(const char *name, int *widthPtr, int *heightPtr, int *pitchPtr) {
	unsigned char *out = convertBitmap(loadPng(name, widthPtr, heightPtr, pitchPtr), 32, *widthPtr, *heightPtr, *pitchPtr, 8);
	*pitchPtr = *widthPtr;
	return out;
}

class SDLCaffeine : public Caffeine {
private:
	unsigned char *titleImage;
	unsigned char *fontImage;
	unsigned char *bumpImage;
	unsigned char *textImage;
public:
	SDLCaffeine(int w, int h, int b, int p) : Caffeine(w, h, b, p) {}
	unsigned char *loadTitle(int *widthPtr, int *heightPtr, int *pitchPtr) {
		titleImage = loadPng8("title-scale.png", widthPtr, heightPtr, pitchPtr);
		return titleImage;
	}
	unsigned char *loadFont(int *widthPtr, int *heightPtr, int *pitchPtr) {
		fontImage = convertBitmap(loadPng("font.png", widthPtr, heightPtr, pitchPtr), 32, *widthPtr, *heightPtr, *pitchPtr, BPP);
		*pitchPtr = ((BPP+7)/8)*(*widthPtr);
		return fontImage;
	}
	unsigned char *loadBumpMap(int *widthPtr, int *heightPtr, int *pitchPtr, bool tileable) {
		if (tileable)
			bumpImage = loadPng8("bumpmap-tileable.png", widthPtr, heightPtr, pitchPtr);
		else
			bumpImage = loadPng8("bumpmap-bg.png", widthPtr, heightPtr, pitchPtr);
		return bumpImage;
	}
	unsigned char *loadText(int *widthPtr, int *heightPtr, int *pitchPtr) {
		textImage = loadPng8("bumpmap-text.png", widthPtr, heightPtr, pitchPtr);
		return textImage;
	}
	void freeTitle() { free(titleImage); }
	void freeFont() { free(fontImage); }
	void freeBumpMap() { free(bumpImage); }
	void freeText() { free(textImage); }
	FILE *loadMod() {
		return fopen("caffeine.mod", "rb");
	}
	bool initAudio() {
		return init_audio();
	}
	void startAudio(void (*callback)(int orderNr, void *data)) {
		player = new ModPlayer(mod, channels, samplerate, bitdepth, sig, littleendian);
		player->addPatternCallback(callback, this);
		start_thread(player);
	}
	void stopAudio() {
		break_play();
		join_thread();
	}
};

int main(int argc, char *argv[]) {
#ifdef __APPLE__
	CFURLRef url = CFBundleCopyBundleURL(CFBundleGetMainBundle());
	char path[1000];
	CFURLGetFileSystemRepresentation(url, true, (UInt8*) path, sizeof(path));
	strcat(path, "/Contents/Resources");
	chdir(path);
	CFRelease(url);
#endif

	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "%s\n", SDL_GetError());
		return 1;
	}
	if (atexit(SDL_Quit) < 0) {
		fprintf(stderr, "atexit(...)\n");
		SDL_Quit();
		return 1;
	}

	int bpp = BPP;

	SDL_WM_SetCaption("Caffeine", "Caffeine");
	screen = SDL_SetVideoMode(WIDTH, HEIGHT, 32, SDL_HWSURFACE | SDL_DOUBLEBUF);
	if (screen == NULL) {
		fprintf(stderr, "Can't set video mode (%s)\n", SDL_GetError());
		return 1;
	}
	SDL_ShowCursor(SDL_DISABLE);
	SDL_Delay(750);
	if (bpp == 12)
		sdl_buffer = SDL_CreateRGBSurface(SDL_SWSURFACE, WIDTH, HEIGHT, 12, 0x0F00, 0x00F0, 0x000F, 0);
	else if (bpp == 16)
		sdl_buffer = SDL_CreateRGBSurface(SDL_SWSURFACE, WIDTH, HEIGHT, 12, 0xF800, 0x07E0, 0x001F, 0);
	else
		sdl_buffer = SDL_CreateRGBSurface(SDL_SWSURFACE, WIDTH, HEIGHT, 32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0);

	if (sdl_buffer == NULL) {
		fprintf(stderr, "Can't create video buffer (%s)\n", SDL_GetError());
		return 1;
	}

	Caffeine *caffeine = new SDLCaffeine(WIDTH, HEIGHT, bpp, sdl_buffer->pitch);

	int startTime = SDL_GetTicks();
	while (true) {
		SDL_Event event;
		bool done = false;
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_KEYDOWN:
			case SDL_QUIT:
				done = true;
				break;
			}
		}
		if (done) break;
		int curTime = SDL_GetTicks() - startTime;

		caffeine->update(curTime);

		caffeine->draw(lock());

		unlockFlip();

		if (caffeine->isFinished())
			break;

		SDL_Delay(1);
	}

	caffeine->stopAudio();
	delete caffeine;

	SDL_FreeSurface(sdl_buffer);

	return 0;
}

