#include <stdlib.h>
#include <math.h>
#include "neon.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define HYPOT(x,y) sqrt((x)*(x) + (y)*(y))

Neon::Neon(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
	sintable = NULL;
	disttable = NULL;
}

Neon::~Neon(void) {
	free(sintable);
	free(disttable);
}

void Neon::init(void) {
	int i;
//	xvar = (512-width)/2;
//	yvar = (512-height)/2;
	xvar = (512 - 320)/2;
	yvar = (512 - 240)/2;

	halfsizeX = 256*width/320;
	halfsizeY = 256*height/240;
	distsizeX = 2*halfsizeX;
	distsizeY = 2*halfsizeY;

	float distscaling = 5;
	if (width > height)
		distscaling = 5.*320/width;
	else
		distscaling = 5.*240/height;

	sintable = (unsigned int*) malloc(sizeof(int)*512);
	for (i = 0; i < 512; i++)
		sintable[i] = (unsigned char) (128+127*sin(2*M_PI*i/512));

	disttable = (unsigned int*) malloc(sizeof(int)*distsizeX*distsizeY);
	unsigned int *ptr = disttable;
	for (int y = 0; y < distsizeY; y++)
		for (int x = 0; x < distsizeX; x++)
			*ptr++ = (int)((1<<8)*distscaling*HYPOT(x - halfsizeX, y - halfsizeY));

	unsigned int *palette = createPalette();
	for (i = 0; i < 512; i++)
		sintable[i] = palette[sintable[i]];
	free(palette);
}

unsigned int *Neon::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	int i;

	for (i = 0; i < 128; i++)
		*ptr++ = 128+i;
	for (i = 128; i < 192; i++)
		*ptr++ = ((i-128)<<2)<<8 | 255;
	for (i = 192; i < 256; i++)
		*ptr++ = ((i-192)<<2)<<16 | 255<<8 | 255;

	if (bpp == 12) {
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>12) & 0x0F00) | ((pal[i]>>8) & 0x00F0) | ((pal[i]>>4) & 0x000F);
	} else if (bpp == 16) {
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>8) & 0xF800) | ((pal[i]>>5) & 0x07E0) | ((pal[i]>>3) & 0x001F);
	}

	return pal;
}

void Neon::draw(unsigned char *buffer, int time) {
	int offsetX1 = (int) (xvar+xvar*sin((double)(3950+time)/900));
	int offsetY1 = (int) (yvar+yvar*cos((double)(2650+time)/750));
	int offsetX2 = (int) (xvar+xvar*sin((double)(4900+time)/850));
	int offsetY2 = (int) (yvar+yvar*cos((double)(1250+time)/1000));
	int offsetX3 = (int) (xvar+xvar*sin((double)(250+time)/750));
	int offsetY3 = (int) (yvar+yvar*cos((double)(2550+time)/950));
	int offsetX4 = (int) (xvar+xvar*sin((double)(2200+time)/1200));
	int offsetY4 = (int) (yvar+yvar*cos((double)(500+time)/1550));

	int noffsetX1 = halfsizeX - (256 - offsetX1)*width/320;
	int noffsetX2 = halfsizeX - (256 - offsetX2)*width/320;
	int noffsetX3 = halfsizeX - (256 - offsetX3)*width/320;
	int noffsetX4 = halfsizeX - (256 - offsetX4)*width/320;
	int noffsetY1 = halfsizeY - (256 - offsetY1)*height/240;
	int noffsetY2 = halfsizeY - (256 - offsetY2)*height/240;
	int noffsetY3 = halfsizeY - (256 - offsetY3)*height/240;
	int noffsetY4 = halfsizeY - (256 - offsetY4)*height/240;

	unsigned int *ptr1, *ptr2, *ptr3, *ptr4;

	if (bpp <= 16) {
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			ptr1 = disttable + ((y + noffsetY1) * distsizeX) + noffsetX1;
			ptr2 = disttable + ((y + noffsetY2) * distsizeX) + noffsetX2;
			ptr3 = disttable + ((y + noffsetY3) * distsizeX) + noffsetX3;
			ptr4 = disttable + ((y + noffsetY4) * distsizeX) + noffsetX4;
			for (int x = 0; x < width; x++)
				*outptr++ = sintable[((*ptr1++ + *ptr2++ + *ptr3++ - *ptr4++) >> 8) & 0x01FF];
			outptr += spitch - width;
		}
	} else {
		unsigned int *outptr = (unsigned int*) buffer;
		int ipitch = pitch/4;
		for (int y = 0; y < height; y++) {
			ptr1 = disttable + ((y + noffsetY1) * distsizeX) + noffsetX1;
			ptr2 = disttable + ((y + noffsetY2) * distsizeX) + noffsetX2;
			ptr3 = disttable + ((y + noffsetY3) * distsizeX) + noffsetX3;
			ptr4 = disttable + ((y + noffsetY4) * distsizeX) + noffsetX4;
			for (int x = 0; x < width; x++)
				*outptr++ = sintable[((*ptr1++ + *ptr2++ + *ptr3++ - *ptr4++) >> 8) & 0x01FF];
			outptr += ipitch - width;
		}
	}
}


