#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "showtext.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

ShowText::ShowText(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;

	font = NULL;
	fontChars = NULL;
	lowerCase = false;
	charWidth = charHeight = 0;
	fontImageWidth = 0;
	plain = true;
	color = 0x00FFFFFF;
	texts = NULL;
	textNum = 0;
	bufferWidth = bufferHeight = 0;
	textTime = startTime = 0;
}

ShowText::~ShowText(void) {
	if (texts==NULL) return;
	unsigned char **ptr = texts;
	unsigned char *b;
	while ((b = *ptr++) != NULL)
		free(b);
	free(texts);
}

void ShowText::init(void) {
}

void ShowText::setFont(unsigned char *f, char *chars, bool lc, int cw, int ch, int w, bool p, unsigned int c) {
	font = f;
	fontChars = chars;
	lowerCase = lc;
	charWidth = cw;
	charHeight = ch;
	fontImageWidth = w;
	plain = p;
	color = c;
	if (bpp == 12)
		color = ((color >> 12) & 0x0F00) | ((color >> 8) & 0x00F0) | ((color >> 4) & 0x000F);
	else if (bpp == 16)
		color = ((color >> 8) & 0xF800) | ((color >> 5) & 0x07E0) | ((color >> 3) & 0x001F);
}

void ShowText::setText(char **text, int time) {
	int maxChars = 0;
	int maxRows = 0;

	char **ptr = text;
	char *s;
	int chars, rows;
	int n = 0;
	while ((s=*ptr++)!=NULL) {
		n++;
		chars = 0;
		rows = 1;

		while (*s!='\0') {
			if (*s=='\n') {
				rows++;
				if (chars>maxChars) maxChars = chars;
				chars = 0;
			} else {
				chars++;
			}
			s++;
		}
		if (chars>maxChars) maxChars = chars;
		if (rows>maxRows) maxRows = rows;
	}

	bufferWidth = maxChars*charWidth;
	bufferHeight = maxRows*charHeight;

	int bytes = (bpp + 7)/8;
	texts = (unsigned char**) malloc(sizeof(char*)*(n+1));
	ptr = text;
	int i = 0;
	while ((s=*ptr++)!=NULL) {
		texts[i] = (unsigned char*) malloc(bytes*bufferWidth*bufferHeight);
		memset(texts[i], 0, bytes*bufferWidth*bufferHeight);
		drawString(texts[i], s);
		i++;
	}
	texts[i] = NULL;
	textNum = n;

	textTime = time/n;

	startTime = -1;
}

void ShowText::drawString(unsigned char *buffer, char *string) {
	int rows = 1;
	char *ptr = string;
	while (*ptr!='\0') {
		if (*ptr=='\n') rows++;
		ptr++;
	}

	int y = (bufferHeight-rows*charHeight)/2;

	char *tempstring = (char*) malloc(strlen(string)+1);
	ptr = string;
	char *outptr = tempstring;
	while (*ptr!='\0') {
		if (*ptr=='\n') {
			*outptr = '\0';
			drawRow(buffer,tempstring,y);
			outptr = tempstring;
			y += charHeight;
		} else {
			*outptr++ = *ptr;
		}
		ptr++;
	}
	*outptr = '\0';
	drawRow(buffer,tempstring,y);
	free(tempstring);
}

void ShowText::drawRow(unsigned char *buffer, char *row, int y) {
	int length = strlen(row);
	int x = (bufferWidth-length*charWidth)/2;
	for (int i = 0; i < length; i++) {
		drawChar(buffer,*row++,x,y);
		x += charWidth;
	}
}

int ShowText::getIndex(char c) {
	if (!lowerCase) c = toupper(c);
	char *pos = strchr(fontChars,c);
	if (pos==NULL) return -1;
	return (pos-fontChars)/sizeof(char);
}

void ShowText::drawChar(unsigned char *buffer, char c, int bufferX, int bufferY) {
	int index = getIndex(c);

	int fontX = charWidth*(index%fontImageWidth);
	int fontY = charHeight*(index/fontImageWidth);

	int fontColor;
	if (bpp <= 16) {
		unsigned short *outptr;
		unsigned short *inptr;
		for (int y = 0; y < charHeight; y++) {
			outptr = ((unsigned short*)buffer) + (bufferY + y)*bufferWidth + bufferX;
			inptr = ((unsigned short*)font) + (fontY + y)*fontImageWidth*charWidth + fontX;
			for (int x = 0; x < charWidth; x++) {
				fontColor = (*inptr++) & 0xFFFF;
				*outptr++ = fontColor;
			}
		}
	} else {
		unsigned int *outptr;
		unsigned int *inptr;
		for (int y = 0; y < charHeight; y++) {
			outptr = ((unsigned int*)buffer) + (bufferY + y)*bufferWidth + bufferX;
			inptr = ((unsigned int*)font) + (fontY + y)*fontImageWidth*charWidth + fontX;
			for (int x = 0; x < charWidth; x++) {
				fontColor = (*inptr++)&0x00FFFFFF;
				if (fontColor)
					*outptr++ = 0xFF000000 | fontColor;
				else
					*outptr++ = 0x00000000;
			}
		}
	}
}

void ShowText::draw(unsigned char *buffer, int time) {
	if (startTime<0) startTime = time;
	time -= startTime;

	int textNr = time/textTime;
	int textPos = time%textTime;

	if (textNr>=textNum) return;

	int x = (width-bufferWidth)/2;
	int y;

	if (textPos < 20*textTime/100) {
		// Fall
		y = (int) ((height-bufferHeight)/2 - ((height-bufferHeight)/2 + bufferHeight)*cos((M_PI/2)*textPos/(20*textTime/100)));
	} else if (textPos < 30*textTime/100) {
		// Bounce
		y = (int) ((height-bufferHeight)/2 - (0.20*(height-bufferHeight)/2)*sin(M_PI*(textPos-20*textTime/100)/(10*textTime/100)));
	} else if (textPos < 75*textTime/100) {
		y = (height-bufferHeight)/2;
	} else {
		// Fall
		y = (int) (height - (height-(height-bufferHeight)/2)*cos((M_PI/2)*(textPos-75*textTime/100)/(25*textTime/100)));
	}

	drawOverlay(buffer, texts[textNr], x, y);
}

void ShowText::drawOverlay(unsigned char *buffer, unsigned char *overlay, int x, int y) {
	int overlayXStart = 0, overlayYStart = 0;
	int bufferXStart = x, bufferYStart = y;

	if (x<0) {
		overlayXStart = -x;
		bufferXStart = 0;
	}
	if (y<0) {
		overlayYStart = -y;
		bufferYStart = 0;
	}

	int bx, by;
	by = bufferYStart;
	int c;
	if (bpp <= 16) {
		unsigned short *outptr, *inptr;
		int spitch = pitch/2;
		for (int oy = overlayYStart; oy < bufferHeight; oy++) {
			if (by>=height) break;
			bx = bufferXStart;
			outptr = ((unsigned short*)buffer) + by*spitch + bufferXStart;
			inptr = ((unsigned short*)overlay) + oy*bufferWidth + overlayXStart;
			for (int ox = overlayXStart; ox < bufferWidth; ox++) {
				if (bx>=width) break;
				c = (*inptr++)&0x0000FFFF;
				if (c)
					if (plain)
						*outptr = color;
					else
						*outptr = c;
				outptr++;
				bx++;
			}
			by++;
		}
	} else {
		unsigned int *outptr, *inptr;
		int ipitch = pitch/4;
		for (int oy = overlayYStart; oy < bufferHeight; oy++) {
			if (by>=height) break;
			bx = bufferXStart;
			outptr = ((unsigned int*)buffer) + by*ipitch + bufferXStart;
			inptr = ((unsigned int*)overlay) + oy*bufferWidth + overlayXStart;
			for (int ox = overlayXStart; ox < bufferWidth; ox++) {
				if (bx>=width) break;
				c = (*inptr++)&0x00FFFFFF;
				if (c)
					if (plain)
						*outptr = color;
					else
						*outptr = c;
				outptr++;
				bx++;
			}
			by++;
		}
	}
}

