#include <stdlib.h>
#include <math.h>
#include "swirl.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define HYPOT(x,y) sqrt((x)*(x) + (y)*(y))

Swirl::Swirl(int w, int h, int b, int p, int p2) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
	period = p2;
	sintable = NULL;
	angletable = NULL;
	disttable = NULL;
}

Swirl::~Swirl(void) {
	free(sintable);
	free(angletable);
	free(disttable);
}

void Swirl::init(void) {
	int i,x,y;
	trigFact = (float)M_PI/period;

	if (width > height)
		curvyfact = 50*240/height;
	else
		curvyfact = 50*240/width;

	sintable = (unsigned int*) malloc(sizeof(int)*512);
	for (i = 0; i < 512; i++)
		sintable[i] = (int)(128+127*sin(2*M_PI*i/512));

	int middleX = width/2, middleY = height/2;

	angletable = (unsigned int*) malloc(sizeof(int)*width*height);
	unsigned int *ptr = angletable;
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			if (x != middleX || y != middleY)
				*ptr++ = (int)((1<<16)*512*atan2(y-middleY,x-middleX)/(2*M_PI));
			else
				*ptr++ = 0;
		}
	}
	disttable = (unsigned int*) malloc(sizeof(int)*width*height);
	ptr = disttable;
	for (y = 0; y < height; y++)
		for (x = 0; x < width; x++)
			*ptr++ = (int)((1<<16)*HYPOT(x-middleX,y-middleY)/10);

	unsigned int *palette = createPalette();
	for (i = 0; i < 512; i++)
		sintable[i] = palette[sintable[i]];
	free(palette);
}

unsigned int *Swirl::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	int i;

	for (i = 0; i < 128; i++)
		*ptr++ = (128-i)<<1;
	for (i = 128; i < 256; i++)
		*ptr++ = ((i-128)<<1)<<8;

	if (bpp == 12) { 
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>12) & 0x0F00) | ((pal[i]>>8) & 0x00F0) | ((pal[i]>>4) & 0x000F);
	} else if (bpp == 16) {
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>8) & 0xF800) | ((pal[i]>>5) & 0x07E0) | ((pal[i]>>3) & 0x001F); 
	}

	return pal;
}

void Swirl::draw(unsigned char *buffer, int time, int multiplier, bool direction, int turnTime) {
	int curvyness = (int) ((1<<8)*curvyfact*sin(trigFact*time));
	int spin;
	if (!direction)
		spin = (int) (time/1.0);
	else
		spin = (int) ((2*turnTime-time)/1.0);

	unsigned int *angleptr = angletable;
	unsigned int *distptr = disttable;
	if (bpp <= 16) {
		unsigned short *outptr = (unsigned short*) buffer;
		int spitch = pitch/2;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++)
				*outptr++ = sintable[(spin+((multiplier*(*angleptr++)+curvyness*((*distptr++)>>8))>>16))&0x01FF];
			outptr += spitch - width;
		}
	} else {
		unsigned int *outptr = (unsigned int*) buffer;
		int ipitch = pitch/4;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++)
				*outptr++ = sintable[(spin+((multiplier*(*angleptr++)+curvyness*((*distptr++)>>8))>>16))&0x01FF];
			outptr += ipitch - width;
		}
	}
}


