#include "symbiancaffeine.h"
#include <eikapp.h>
#include <eikenv.h>
#include <eikappui.h>
#include <fbs.h>
#include <flogger.h>

#include <caffeine.mbg>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "caffeine.h"
#include "bitmapfuncs.h"

#include "modplayer.h"
#include "wavestream.h"

//#include <settinginfo.h>
#ifdef EKA2
#include <centralrepository.h>
#include <ProfileEngineSDKCRKeys.h>
#endif

class CModGenerator : public CBase, public MWaveGenerator {
public:
	CModGenerator(SymbianCaffeine* aCaffeine, ModFile* aMod) {
		iCaffeine = aCaffeine;
		iMod = aMod;
	}
	void RenderData(TAny* aUserData, TInt16* aPtr, TInt aLength) {
		iPlayer->read((unsigned char*) aPtr, 2*aLength);
	}
	void SetSampleRate(TInt aSampleRate) {
		if (iPlayer == NULL) {
			iPlayer = new ModPlayer(iMod, 1, aSampleRate, 16, true, true);
			iPlayer->addPatternCallback(iCallback, iCaffeine);
			iCaffeine->SetPlayer(iPlayer);
		}
	}
	void SetCallback(void (*callback)(int orderNr, void *data)) {
		iCallback = callback;
	}

private:
	SymbianCaffeine* iCaffeine;
	ModFile* iMod;
	ModPlayer* iPlayer;
	void (*iCallback)(int orderNr, void *data);
};

SymbianCaffeine::SymbianCaffeine(int w, int h, int b, int p) : Caffeine(w, h, b, p) {
	iStream = NULL;
	iGenerator = NULL;
	iMusicStarted = EFalse;

	CApaApplication *app = CEikonEnv::Static()->EikAppUi()->Application();
	TParsePtrC parse(app->AppFullName());
	iBitmapFilename = parse.DriveAndPath();
#ifdef EKA2
	iBitmapFilename.SetLength(2);
	TBuf<50> privPath;
	privPath.Format(_L("\\private\\%08x\\"), app->AppDllUid().iUid);
	iBitmapFilename.Append(privPath);
#endif

	for (int i = 0; i < iBitmapFilename.Length(); i++)
		iModFilename[i] = iBitmapFilename[i];
	iModFilename[iBitmapFilename.Length()] = '\0';

	strcat(iModFilename, "caffeine.mod");

	_LIT(KMBMFile, "caffeine.mbm");
	iBitmapFilename.Append(KMBMFile);
}

SymbianCaffeine::~SymbianCaffeine() {
	CloseSTDLIB();
}

unsigned char *SymbianCaffeine::LoadBitmapL(CFbsBitmap*& aBitmap, TInt aId, int *widthPtr, int *heightPtr, int *pitchPtr) {
	aBitmap = new(ELeave) CFbsBitmap();
	User::LeaveIfError(aBitmap->Load(iBitmapFilename, aId));
	*widthPtr = aBitmap->SizeInPixels().iWidth;
	*heightPtr = aBitmap->SizeInPixels().iHeight;
	*pitchPtr = aBitmap->ScanLineLength(*widthPtr, aBitmap->DisplayMode());
#ifdef EKA2
	aBitmap->LockHeap();
#endif
	return (unsigned char*) aBitmap->DataAddress();
}

void SymbianCaffeine::FreeBitmap(CFbsBitmap*& aBitmap) {
#ifdef EKA2
	aBitmap->UnlockHeap();
#endif
	delete aBitmap;
	aBitmap = NULL;
}

unsigned char *SymbianCaffeine::loadTitle(int *widthPtr, int *heightPtr, int *pitchPtr) {
	return LoadBitmapL(iTitle, EMbmCaffeineTitle_scale, widthPtr, heightPtr, pitchPtr);
}

void SymbianCaffeine::freeTitle() {
	FreeBitmap(iTitle);
}

unsigned char *SymbianCaffeine::loadFont(int *widthPtr, int *heightPtr, int *pitchPtr) {
	CFbsBitmap *bitmap = new(ELeave) CFbsBitmap();
	User::LeaveIfError(bitmap->Load(iBitmapFilename, EMbmCaffeineFont));
	*widthPtr = bitmap->SizeInPixels().iWidth;
	*heightPtr = bitmap->SizeInPixels().iHeight;
	*pitchPtr = bitmap->ScanLineLength(*widthPtr, bitmap->DisplayMode());
#ifdef EKA2
	bitmap->LockHeap();
#endif
	iFont = convertBitmap((unsigned char*) bitmap->DataAddress(), 8, *widthPtr, *heightPtr, *pitchPtr, bpp, false);
#ifdef EKA2
	bitmap->UnlockHeap();
#endif
	delete bitmap;
	*pitchPtr = ((bpp + 7)/8)*(*widthPtr);
	return iFont;
}

void SymbianCaffeine::freeFont() {
	free(iFont);
}

unsigned char *SymbianCaffeine::loadBumpMap(int *widthPtr, int *heightPtr, int *pitchPtr, bool tileable) {
	if (tileable)
		return LoadBitmapL(iBumpMap, EMbmCaffeineBumpmap_tileable, widthPtr, heightPtr, pitchPtr);
	else
		return LoadBitmapL(iBumpMap, EMbmCaffeineBumpmap_bg, widthPtr, heightPtr, pitchPtr);
}

void SymbianCaffeine::freeBumpMap() {
	FreeBitmap(iBumpMap);
}

unsigned char *SymbianCaffeine::loadText(int *widthPtr, int *heightPtr, int *pitchPtr) {
	return LoadBitmapL(iText, EMbmCaffeineBumpmap_text, widthPtr, heightPtr, pitchPtr);
}

void SymbianCaffeine::freeText() {
	FreeBitmap(iText);
}

FILE *SymbianCaffeine::loadMod() {
	return fopen(iModFilename, "rb");
}

TInt SymbianCaffeine::GetVolume() {
	TInt volume = 50;
	TInt ringing = 0;
	TInt err = KErrNone;

#ifdef EKA2
	CRepository *cr = CRepository::NewLC(KCRUidProfileEngine);
	err = cr->Get(KProEngActiveRingingType, ringing);
	CleanupStack::PopAndDestroy(cr);
#else

#if !(defined(__WINS__) || defined(__WINSCW__))
/*	TRAPD(err, 
		CSettingInfo *info = CSettingInfo::NewL(NULL);
		CleanupStack::PushL(info);
		User::LeaveIfError(info->Get(SettingInfo::ERingingType, ringing));
		CleanupStack::PopAndDestroy();
		);
*/

	RLibrary dll;
	err = dll.Load(_L("platformenv.dll"), TUidType());
	if (err == KErrNone) {
		typedef CBase* (*InfoNewLPtr)(TAny*);
		typedef TInt (*InfoGetPtr)(CBase*, TInt, TInt&);
		InfoNewLPtr NewLPtr;
		InfoGetPtr GetPtr;
		// winscw
//		NewLPtr = (InfoNewLPtr) dll.Lookup(21);
//		GetPtr = (InfoGetPtr) dll.Lookup(19); // doesn't really work, uses another calling convention
		// armi
		NewLPtr = (InfoNewLPtr) dll.Lookup(20);
		GetPtr = (InfoGetPtr) dll.Lookup(19);

		CBase *info = NewLPtr(NULL);
		CleanupStack::PushL(info);
		err = GetPtr(info, 1, ringing); // SettingInfo::ERingingType
		CleanupStack::PopAndDestroy();
		dll.Close();
	}
#endif

#endif

	if (err == KErrNone) {
		if (ringing == 3) volume = 13;
		else if (ringing == 4) volume = 0;
	}
	return volume;
}

bool SymbianCaffeine::initAudio() {
	iGenerator = new(ELeave) CModGenerator(this, mod);
	iStream = CWaveStream::NewL(_L("Caffeine sound"), iGenerator, GetVolume());
	return true;
}

void SymbianCaffeine::startAudio(void (*callback)(int orderNr, void *data)) {
	iGenerator->SetCallback(callback);
	iStream->Start();
	iMusicStarted = ETrue;
}

void SymbianCaffeine::stopAudio() {
	if (iStream)
		iStream->Stop();
	delete iStream;
	iStream = NULL;
	delete iGenerator;
	iGenerator = NULL;
}

void SymbianCaffeine::SetPlayer(ModPlayer* aPlayer) {
	player = aPlayer;
}

void SymbianCaffeine::EnableMusic(TBool aEnable) {
	if (!iMusicStarted)
		return;
	if (aEnable)
		iStream->Start();
	else
		iStream->Stop();
}


